package com.glintt.at.dmis.ws.core;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateException;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.KeyGenerator;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.SecretKeySpec;

/**
 * @author Ricardo Rodrigues (created based on the applet source code available on
 *         https://faturas.portaldasfinancas.gov.pt/testarLigacaoWebService.action )
 */
public final class CypherUtils {

    private CypherUtils() {
    }

    private enum AESConfiguration {
        CIPHER_ALGORITHM("AES"), CIPHER_MODE("ECB"), CIPHER_PADDING("PKCS5Padding");

        private String value;

        private AESConfiguration(final String value) {
            this.value = value;
        }

        public String getValue() {
            return this.value;
        }
    }

    private enum RSAConfiguration {
        CIPHER_ALGORITHM("RSA");

        private String value;

        private RSAConfiguration(final String value) {
            this.value = value;
        }

        public String getValue() {
            return this.value;
        }
    }

    enum KSConfiguration {
        KEYSTORE_TYPE("JKS");

        private String value;

        KSConfiguration(final String value) {
            this.value = value;
        }

        public String getValue() {
            return this.value;
        }
    }

    public static Key getPublicKeyFromKeystore(final String keystorePath, final String keystorePwd,
            final String keyAlias)
            throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {

        final KeyStore ks = KeyStore.getInstance(KSConfiguration.KEYSTORE_TYPE.getValue());

        final InputStream fis = CypherUtils.class.getClassLoader().getResourceAsStream(keystorePath);
        ks.load(fis, keystorePwd.toCharArray());
        final Key key = ks.getCertificate(keyAlias).getPublicKey();

        fis.close();

        return key;
    }

    public static byte[] cypherRequestKey(final PublicKey publicKey, final byte[] requestKey)
            throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException,
            IllegalBlockSizeException, BadPaddingException, UnsupportedEncodingException {
        final Cipher cipher = Cipher.getInstance(RSAConfiguration.CIPHER_ALGORITHM.getValue());
        cipher.init(Cipher.ENCRYPT_MODE, publicKey);
        return cipher.doFinal(requestKey);
    }

    public static byte[] decypherRequestKey(final PrivateKey privateKey,
            final byte[] encriptedRequestKey) throws NoSuchAlgorithmException, NoSuchPaddingException,
            InvalidKeyException, IllegalBlockSizeException, BadPaddingException {
        byte[] decryptedMessage;

        final Cipher cipher = Cipher.getInstance(RSAConfiguration.CIPHER_ALGORITHM.getValue());
        cipher.init(Cipher.DECRYPT_MODE, privateKey);

        decryptedMessage = cipher.doFinal(encriptedRequestKey);

        return decryptedMessage;
    }

    public static byte[] cypherCredential(final byte[] requestKey, final String credential)
            throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException,
            InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException,
            UnsupportedEncodingException {
        final Cipher cipher = Cipher.getInstance(AESConfiguration.CIPHER_ALGORITHM.getValue() + "/"
                + AESConfiguration.CIPHER_MODE.getValue() + "/"
                + AESConfiguration.CIPHER_PADDING.getValue());

        cipher.init(Cipher.ENCRYPT_MODE,
                new SecretKeySpec(requestKey, AESConfiguration.CIPHER_ALGORITHM.getValue()));

        return cipher.doFinal(credential.getBytes("UTF-8"));
    }

    public static byte[] cypherCredentialBuffer(final byte[] requestKey, final byte[] credential)
            throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException,
            InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException,
            UnsupportedEncodingException {
        final Cipher cipher = Cipher.getInstance(AESConfiguration.CIPHER_ALGORITHM.getValue() + "/"
                + AESConfiguration.CIPHER_MODE.getValue() + "/"
                + AESConfiguration.CIPHER_PADDING.getValue());

        cipher.init(Cipher.ENCRYPT_MODE,
                new SecretKeySpec(requestKey, AESConfiguration.CIPHER_ALGORITHM.getValue()));

        return cipher.doFinal(credential);
    }

    public static String decypherCredential(final byte[] requestKey, final byte[] encriptedCredential)
            throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException,
            InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException,
            UnsupportedEncodingException {

        final Cipher decipher = Cipher.getInstance(AESConfiguration.CIPHER_ALGORITHM.getValue() + "/"
                + AESConfiguration.CIPHER_MODE.getValue() + "/"
                + AESConfiguration.CIPHER_PADDING.getValue());

        decipher.init(Cipher.DECRYPT_MODE,
                new SecretKeySpec(requestKey, AESConfiguration.CIPHER_ALGORITHM.getValue()));

        return new String(decipher.doFinal(encriptedCredential), "UTF-8");
    }

    public static byte[] decypherCredentialBuffer(final byte[] requestKey,
            final byte[] encriptedCredential) throws NoSuchAlgorithmException, NoSuchPaddingException,
            InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException,
            BadPaddingException, UnsupportedEncodingException {

        final Cipher decipher = Cipher.getInstance(AESConfiguration.CIPHER_ALGORITHM.getValue() + "/"
                + AESConfiguration.CIPHER_MODE.getValue() + "/"
                + AESConfiguration.CIPHER_PADDING.getValue());

        decipher.init(Cipher.DECRYPT_MODE,
                new SecretKeySpec(requestKey, AESConfiguration.CIPHER_ALGORITHM.getValue()));

        return decipher.doFinal(encriptedCredential);
    }

    public static byte[] generateRequestKey() throws NoSuchAlgorithmException {
        final KeyGenerator generator = KeyGenerator
                .getInstance(AESConfiguration.CIPHER_ALGORITHM.getValue());
        generator.init(128);

        final Key keyToBeWrapped = generator.generateKey();

        return keyToBeWrapped.getEncoded();
    }

}
