package com.glintt.at.dmis.ws.core;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;
import java.util.logging.Logger;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import javax.xml.soap.SOAPHeader;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import sun.misc.BASE64Encoder;

/**
 * @author Ricardo Rodrigues (created based on the applet source code available on
 *         https://faturas.portaldasfinancas.gov.pt/testarLigacaoWebService.action )
 */
public class WSSHeaderHandler implements SOAPHandler<SOAPMessageContext> {

    private static final String AUTH_NS = "http://schemas.xmlsoap.org/ws/2002/12/secext";
    private static final String AUTH_PREFIX = "wss";

    private static final String KS_PATH = "saPubKey.jks";
    private static final String KS_PWD = "saKeyPubPass";
    private static final String KEY_ALIAS = "sapubkey.prod";

    private final Logger logger;
    private final String userName;
    private final String password;

    public WSSHeaderHandler(final Logger logger, final String userName, final String password) {
        this.logger = logger;
        this.userName = userName;
        this.password = password;
    }

    public boolean handleFault(final SOAPMessageContext smc) {
        return true;
    }

    public void close(final MessageContext mc) {
    }

    /**
     * adiciona header para autenticacao
     */
    public boolean handleMessage(final SOAPMessageContext smc) {

        final boolean isRequest = (Boolean) smc.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);

        if (isRequest) {
            try {
                final Key publicKey = CypherUtils.getPublicKeyFromKeystore(WSSHeaderHandler.KS_PATH,
                        WSSHeaderHandler.KS_PWD, WSSHeaderHandler.KEY_ALIAS);

                // ----------------------
                // ---- AUTENTICACAO ----
                // ----------------------
                // Generate simetric key used for this request. Nonce!
                final byte[] simetricKey = WSSHeaderHandler.generateAESKey();
                // logger.info("Chave sim�trica gerada");

                // Encrypt with the simetric key and B64 encode the password
                final byte[] encryptedPassword = CypherUtils.cypherCredential(simetricKey,
                        this.password);
                final String b64EncryptedPassword = new BASE64Encoder().encodeBuffer(encryptedPassword);
                // logger.info("Password field encrypted and encoded: " +
                // b64EncryptedPassword);

                // Encrypt with the simetric key and B64 encode the timestamp
                final String timestamp = WSSHeaderHandler.getTimestamp();
                final byte[] encryptedTimestamp = CypherUtils.cypherCredential(simetricKey, timestamp);
                final String b64EncryptedTimestamp = new BASE64Encoder()
                        .encodeBuffer(encryptedTimestamp);
                // logger.info("timestamp encrypted and encoded: " +
                // b64EncryptedTimestamp);

                // Encrypt with the SA public key and B64 encode the request simetric key
                // (nonce)
                final byte[] encriptedSimetricKey = CypherUtils.cypherRequestKey((PublicKey) publicKey,
                        simetricKey);
                final String b64EncryptedSimetricKey = new BASE64Encoder()
                        .encodeBuffer(encriptedSimetricKey);
                // logger.info("Request simetric key encrypted and encoded:\n" +
                // b64EncryptedSimetricKey);
                // ----------------------

                final SOAPEnvelope envelope = smc.getMessage().getSOAPPart().getEnvelope();
                final SOAPFactory soapFactory = SOAPFactory.newInstance();

                final SOAPElement wsSecHeaderElm = this.createSecurityHeader(soapFactory, this.userName,
                        b64EncryptedPassword, b64EncryptedSimetricKey, b64EncryptedTimestamp);

                // create SOAPHeader instance for SOAP envelope
                final SOAPHeader soapHeader = envelope.addHeader();

                // add SOAP element for header to SOAP header object
                soapHeader.addChildElement(wsSecHeaderElm);

            } catch (final Exception ex) {
                final StringWriter sw = new StringWriter();
                ex.printStackTrace(new PrintWriter(sw));
                final String stacktrace = sw.toString();
                this.logger.severe("Could not intercept and add WS Security Header\n\n" + stacktrace);
            }
        }

        return true;
    }

    private SOAPElement createSecurityHeader(final SOAPFactory soapFactory, final String username,
            final String password, final String simetricKey, final String timestamp)
            throws SOAPException {
        // WSSecurity <Security> header
        final SOAPElement wsSecHeaderElm = soapFactory.createElement("Security",
                WSSHeaderHandler.AUTH_PREFIX, WSSHeaderHandler.AUTH_NS);

        final SOAPElement userNameTokenElm = soapFactory.createElement("UsernameToken",
                WSSHeaderHandler.AUTH_PREFIX, WSSHeaderHandler.AUTH_NS);
        // Username
        final SOAPElement userNameElm = soapFactory.createElement("Username",
                WSSHeaderHandler.AUTH_PREFIX, WSSHeaderHandler.AUTH_NS);
        userNameElm.addTextNode(username);

        userNameTokenElm.addChildElement(userNameElm);
        if (password != null && !password.isEmpty()) {
            // Password
            final SOAPElement passwdElm = soapFactory.createElement("Password",
                    WSSHeaderHandler.AUTH_PREFIX, WSSHeaderHandler.AUTH_NS);
            passwdElm.addTextNode(password);

            // Nonce
            final SOAPElement nonceElm = soapFactory.createElement("Nonce", WSSHeaderHandler.AUTH_PREFIX,
                    WSSHeaderHandler.AUTH_NS);
            nonceElm.addTextNode(simetricKey);
            // Created
            final SOAPElement createdElm = soapFactory.createElement("Created",
                    WSSHeaderHandler.AUTH_PREFIX, WSSHeaderHandler.AUTH_NS);
            createdElm.addTextNode(timestamp);

            userNameTokenElm.addChildElement(passwdElm);
            userNameTokenElm.addChildElement(nonceElm);
            userNameTokenElm.addChildElement(createdElm);
        }

        // add child elements to the root element
        wsSecHeaderElm.addChildElement(userNameTokenElm);

        return wsSecHeaderElm;
    }

    public java.util.Set<QName> getHeaders() {
        return null;
    }

    // ---------------
    // Utility Methods
    // ---------------

    private static byte[] generateAESKey() throws NoSuchAlgorithmException {
        return CypherUtils.generateRequestKey();
    }

    private static final SimpleDateFormat TIMESTAMP_FORMATER = new SimpleDateFormat(
            "yyyy-MM-dd'T'HH:mm:ss.S'Z'");

    static {
        WSSHeaderHandler.TIMESTAMP_FORMATER.setTimeZone(TimeZone.getTimeZone("UTC"));
    }

    private static String getTimestamp() throws ParseException {
        final Calendar c = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        return WSSHeaderHandler.TIMESTAMP_FORMATER.format(c.getTime());
    }
}